# bash range expand library 0.12
# (C) Copyright  Heinz Deinhart 2015-2024
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This is a bash library, it has to be sourced not called.
#
# bash-range-expand version 0.12

#@@ error messages

#@ error-codes (bash-csv)
# <50 .. reserved for base-base error-codes
#  51 .. illegal handle-name
#  52 .. handle not loaded
#  53 .. handle already in use
#  9x .. errors local to each function


[ "$0" != "${BASH_SOURCE[0]}" ] || {
	echo "$0: this script must be sourced, not called" >&2
	exit 1
}

#@@ range expansion

#@ ranges
# a range consist of one or more separated by separator
#  + <aliase> (^[[:alpha:]][[:alnum:]_]*$) references another range
#  + <name> (^[[:alpha:]][[:alnum:]_]*$) expands to itself
#  + <number> expands to <prefix><number><suffix>
#  * <number>-<higher_number> expands to all numbers in that range
# a separator must not contain the following chars: [[:alnum:]_]
# default separator is space (' ')
# each alias can set its own prefix and suffix
# if none is defined in an alias the globally set prefix and suffix will be used
# default prefix and suffix is ""

#@@ range expand

#@ bre_expand <ret-array-vn> <handle> <range> [..]
# expand ranges found in loaded handle and return it in ret-array
# range-dict keys are names of ranges (whitespace not allowed)
#   and their values can be:
#     a macro: any text string that matches a key of range-dict
#              its value will be expanded in recursive fashion
#     a range (<number>-<number>): will be expanded to each member of range
#                                  prefix and suffix will be added to each
#     a number: prefix and suffix will be added
#     any other string will be copied unmodified
# errors: 2, 3, 91 (illegal range), 92 (reffered alias not found),
#         93 (ref loop detected), 94 (multiple errors happened)
function bre_expand() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 more args"
	return 2
    }

    _bre_expand_x full "$@" || return "$?"
}

#@ bre_expand <ret-array-vn> <handle> <range> [..]
# works like bre_expand but stop after one expansion, no recursion
# errors: 2, 3, 91 (illegal range), 92 (reffered alias not found),
#         93 (ref loop detected), 94 (multiple errors happened)
function bre_expand_shallow() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 more args"
	return 2
    }

    _bre_expand_x shallow "$@" || return "$?"
}

#@@ handle helpers

#@ bre_handle_is_loaded <handle>
# check if handle is open
# errors: 2, 51, 52
function bre_handle_is_loaded() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    bb_vn_validate "$1" || return 51
    bb_dict_is_set "bre_hnd__$1_dict" || return 52
}

#@ bre_handle_create_empty <handle>
# create an empty handle
# errors: 1, 2, 51, 52
function bre_handle_create_empty() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    if bre_handle_is_loaded "$1"
    then
	return 53 # handle already in use
    else
	case "$?" in
	    2) return 2;; # internal error: illegal arg count
	    51) return 51;; # illegal chars in handle
	    52) ;; # not loaded
	    *) return 1
	esac
    fi

    declare -A -g "bre_hnd__${1}_dict"
    declare -g "bre_hnd__${1}_global_prefix"
    declare -g "bre_hnd__${1}_global_suffix"
    local -n _dict_ref="bre_hnd__${1}_dict"
    _dict_ref=()
}

#@ bre_handle_create_from_ini <handle> <path>
# errors: 51, 52, 53, 91 (failed to load ini from path), 92 (bash-ini error)
# shellcheck disable=SC2178
function bre_handle_create_from_ini() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _handle="$1"
    local _path="$2"

    local _dict_vn="bre_hnd__${_handle}_dict"
    local _global_prefix_vn="bre_hnd__${_handle}_global_prefix"
    local _global_suffix_vn="bre_hnd__${_handle}_global_suffix"
    local _db="bre_load_from_ini__inidb"

    declare -A -g "$_dict_vn"
    declare -g "$_global_prefix_vn"
    declare -g "$_global_suffix_vn"

    local -i _retval

    # make sure handle is not already in use
    if bre_handle_is_loaded "$_handle"
    then
	return 53
    else
	_retval="$?"
	case "$_retval" in
	    52) ;;
	    *) return "$_retval"
	esac
    fi

    bini_db_load "$_db" "$_path" || {
	bb_msg warning "bre_load_from_ini: bini_db_load failed: $?"
	return 91
    }

    # get global prefix and suffix
    bini_get_value "$_global_prefix_vn" "$_db" "host_range_prefix" || {
	_retval="$?"
	case "$_retval" in
	    54) ;; # key not found, there is no global setting
	    *) bb_msg err \
		      "bre_load_from_ini: failed to get global prefix: $_retval"
	       bini_db_destroy "$_db" || true
	       return 92
	       ;;
	esac
    }
    bini_get_value "$_global_suffix_vn" "$_db" "host_range_suffix" || {
	_retval="$?"
	case "$_retval" in
	    54) ;; # key not found, there is no global setting
	    *) bb_msg err \
		      "bre_load_from_ini: failed to get global suffix: $_retval"
	       bini_db_destroy "$_db" || true
	       return 92
	       ;;
	esac
    }

    # read aliases
    # shellcheck disable=SC2034 # disable unused warning
    local -a bre_load_from_ini__sections
    bini_get_section_list bre_load_from_ini__sections "$_db" || {
	bb_msg err \
	       "bre_load_from_ini: failed to get section list: $?"
	bini_db_destroy "$_db" || true
	return 92
    }
    local -n _dict_ref="$_dict_vn"
    _dict_ref=()
    bb_array_iter bre_load_from_ini__sections \
		  _bre_load_from_ini_alias_iterator "$_db" "$_dict_vn"

    bini_db_destroy bre_load_from_ini__inidb || {
	bb_msg err "bre_load_from_ini: bini_db_destroy failed: $?"
	return 92
    }
}

#@ bre_discard <handle>
# discard a handle
# errors: 51, 52
function bre_handle_discard() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    local _handle="$1"
    local _dict_vn="bre_hnd__${_handle}_dict"
    local _global_prefix="bre_hnd__${_handle}_global_prefix"
    local _global_suffix="bre_hnd__${_handle}_global_suffix"

    bre_handle_is_loaded "$_handle" || return "$?"

    unset "$_dict_vn" "$_global_prefix" "$_global_suffix"
}

### internal functions below ###

# _bre_expand_alias <mode> <add-to-array-vn> <handle> <range> [<seen-dict-vn>]
# tries to treat and expand range as alias
# note: seen-dict is used in recursive calls to prevent endless ref loops
# 2, 91 (range is no alias), 92 (ref not found), 93 (ref loop occured)
function _bre_expand_alias() {
    (( $# == 4 || $# == 5 )) || {
	bb_ierr "called with $# instead of 4 or 5 args"
	return 2
    }

    local _mode="$1"
    local _target_array_vn="$2"
    local _handle="$3"
    local _range="$4"

    local -i _shallow
    case "$_mode" in
	full) _shallow=0;;
	shallow) _shallow=1;;
	*) bb_ierr "called with illegal mode: $_mode"
	   return 2
	   ;;
    esac

    local _seen_dict_vn
    if (( $# == 4 ))
    then
	# we are called from outside
	# shellcheck disable=SC2034
	local -A bre_expand_alias__seen_ar=(["$_range"]=1)
	_seen_dict_vn=bre_expand_alias__seen_ar
    else
	# this is a recursiv call
	# check for dup to prohibit endless recursion
	_seen_dict_vn="$5"
	bb_assert_dict_is_set "$_seen_dict_vn"
	if bb_dict_contains_key "$_seen_dict_vn" "$_range"
	then
	    return 93
	fi

	# add current alias to seen dict
	local -n _seen_dict_ref="$_seen_dict_vn"
	_seen_dict_ref["$_range"]=1
    fi

    local _dict_vn="bre_hnd__${_handle}_dict"
    local _global_prefix_vn="bre_hnd__${_handle}_global_prefix"
    local _global_suffix_vn="bre_hnd__${_handle}_global_suffix"

    # shellcheck disable=SC2178
    local -n _dict_ref="$_dict_vn"
    local -i _is_alias=0
    local bre_expand_alias__prefix bre_expand_alias__suffix

    # check if range is an alias and got host-ranges
    if bb_dict_contains_key "$_dict_vn" "ranges_$_range"
    then
	bb_dict_get_value bre_expand_alias__prefix "$_dict_vn" \
			  "prefix_$_range" ""
	bb_dict_get_value bre_expand_alias__suffix "$_dict_vn" \
			  "suffix_$_range" ""
	# shellcheck disable=SC2086 # we want word split
	_bre_expand_simple_ranges "$_target_array_vn" \
				  "$bre_expand_alias__prefix" \
				  "$bre_expand_alias__suffix" \
				  ${_dict_ref[ranges_$_range]}
	_is_alias=1
    fi

    local -i _retval=0

    # check if range is an alias and got refs
    if bb_dict_contains_key "$_dict_vn" "refs_$_range"
    then # range is an alias and got refs -> recurse
	local _a
	for _a in ${_dict_ref["refs_$_range"]}
	do
	    if (( _shallow == 0 ))
	    then
		_bre_expand_alias "$_mode" "$_target_array_vn" "$_handle" \
				  "$_a" "$_seen_dict_vn" || _retval=92
	    else
		bb_array_add "$_target_array_vn" "$_a"
	    fi
	done
	_is_alias=1
    fi

    # return 91 if this was not an alias
    (( _is_alias == 1 )) || return 91

    return "$_retval"
}

# _bre_parse_ranges <add-to-array-vn> <prefix> <suffix> <range> [..]
# this funs expends simple ranges, it assumes that a range is no alias
# errors: 2, 91 (illegal range)
function _bre_expand_simple_ranges() {
    (( $# > 3 )) || {
	bb_ierr "called with $# instead of 4 or more args"
	return 2
    }

    local _target_array_vn="$1"
    local _prefix="$2"
    local _suffix="$3"
    shift 2 # 2 not 3 because we shift at begin of while loop

    local -i _i _from _to

    while (( $# > 1 ))
    do
	shift

	# ignore empty range
	[ -n "$1" ] || continue

	# from-to range, add numbers in range with prefix and suffix
	if [[ "$1" =~ ^[[:digit:]]+-[[:digit:]]+$ ]]
	then
	    _from="${1%-*}"
	    _to="${1#*-}"

	    if (( _from > _to ))
	    then
		bb_msg warning "bre_expand: illegal simple range: $1"
		return 91
	    else
		for (( _i=_from; _i <= _to; _i++ ))
		do
		    bb_array_add "$_target_array_vn" "$_prefix$_i$_suffix"
		done
	    fi
	    continue
	fi

	# number only, add it with prefix and suffix
	if [[ "$1" =~ ^[[:digit:]]+$ ]]
	then
	    bb_array_add "$_target_array_vn" "$_prefix$1$_suffix"
	    continue
	fi

	# consider range to be hostname (done add prefix and suffix)
	bb_array_add "$_target_array_vn" "$1"
    done
}

# _bre_load_from_ini_alias_iterator <alias> <ini-handle> <dict-vn>
function _bre_load_from_ini_alias_iterator() {
    (( $# == 3 )) || {
	bb_ierr "called with $# instead of 3 args"
	return 2
    }

    local _alias="$1"
    local _db="$2"
    # shellcheck disable=SC2178
    local -n _dict_ref="$3"

    local -i _retval
    local bre_load_from_ini_alias_iterator__s

    local -i _found_ranges=0
    # first load ranges and refs, so we don't care about suffix and
    # prefix if there are no ranges and no refs
    if bini_get_value bre_load_from_ini_alias_iterator__s "$_db" \
		      "host_ranges" "$_alias"
    then
	_dict_ref["ranges_$_alias"]="$bre_load_from_ini_alias_iterator__s"
	_found_ranges=1
    else
	_retval="$?"
	case "$_retval" in
	    54) ;; # no ranges found
	    *) bb_msg err "_bre_load_from_ini_alias_iterator:" \
		      "failed to get host_range: $_retval"
	       return 92
	esac
    fi
    if bini_get_value bre_load_from_ini_alias_iterator__s "$_db" \
		      "refs" "$_alias"
    then
	_dict_ref["refs_$_alias"]="$bre_load_from_ini_alias_iterator__s"
    else
	_retval="$?"
	case "$_retval" in
	    54) ;; # no refs found
	    *) bb_msg err "_bre_load_from_ini_alias_iterator:" \
		      "failed to get refs: $_retval"
	       return 92
	esac
    fi
    (( _found_ranges > 0 )) || return # no need for pre/suffix when no ranges

    # load prefix and suffix
    if bini_get_value_with_default bre_load_from_ini_alias_iterator__s "$_db" \
				   "host_range_prefix" "$_alias"
    then
	_dict_ref["prefix_$_alias"]="$bre_load_from_ini_alias_iterator__s"
    else
	_retval="$?"
	case "$_retval" in
	    54) ;; # key not found, no big deal
	    *) bb_msg err "_bre_load_from_ini_alias_iterator:" \
		      "failed to get prefix: $_retval"
	       return 92
	esac
    fi
    if bini_get_value_with_default bre_load_from_ini_alias_iterator__s "$_db" \
				   "host_range_suffix" "$_alias"
    then
	_dict_ref["suffix_$_alias"]="$bre_load_from_ini_alias_iterator__s"
    else
	_retval="$?"
	case "$_retval" in
	    54) ;; # key not found, no big deal
	    *) bb_msg err "_bre_load_from_ini_alias_iterator:" \
		      "failed to get suffix: $_retval"
	       return 92
	esac
    fi
}
# bre_expand_x <mode> <ret-array-vn> <handle> <range> [..]
function _bre_expand_x() {
    (( $# > 3 )) || {
	bb_ierr "called with $# instead of 4 more args"
	return 2
    }

    local _mode="$1"
    local _ret_vn="$2"
    local _handle="$3"
    shift 3

    bb_vn_validate "$_ret_vn" || return 3
    bre_handle_is_loaded "$_handle" || return "$?"

    # clean return array
    local -n _ret_ref="$_ret_vn"
    _ret_ref=()

    local -i _retval _err_91=0 _err_92=0 _err_93=0

    # for each specify range, test if it is alias if not use simple expand
    while (( $# > 0 ))
    do
	_bre_expand_alias "$_mode" "$_ret_vn" "$_handle" "$1" || {
	    _retval="$?"

	    if (( _retval == 91 ))
	    then
		# _bre_expand_alias says that range is no alias
		local -n _prefix="bre_hnd__${_handle}_global_prefix"
		local -n _suffix="bre_hnd__${_handle}_global_suffix"
		_bre_expand_simple_ranges "$_ret_vn" "${_prefix-}" \
					  "${_suffix-}" "$1" || {
		    _retval="$?"
		    case "$_retval" in
			0) ;;
			91) _err_91=1;;
			2) return 2;;
			*) bb_fatal \
			       "internal error: _bre_expand_simple_ranges" \
			       "returned $_retval"
		    esac
		}
	    else
		case "$_retval" in
		    92) _err_92=1;;
		    93) _err_93=1;;
		    2) return 2;;
		    *) bb_fatal \
			   "internal error: _bre_expand_alias" \
			   "returned $_retval"
		esac
	    fi
	}
	shift
    done

    case $(( _err_91 + _err_92 + _err_93 )) in
	0) return 0;; # no error \o/
	1) # exactly one error occured
	    (( _err_91 == 0 )) || return 91
	    (( _err_92 == 0 )) || return 92
	    return 93
	    ;;
	*) return 94 # multiple errors occured
    esac
}
